import { MemoizeFunc } from 'fast-memoize';
import objectOmit from 'object.omit';
import objectPick from 'object.pick';
declare type AnyConstructor<GType = unknown> = new (...args: any[]) => GType;
declare type AnyFunction<GType = any> = (...args: any[]) => GType;
interface PlainObject {
    [key: string]: unknown;
}
/**
 * Type cast an argument. If no type is provided it will default to any.
 *
 * @param arg - the arg to typecast
 */
export declare const Cast: <GType = any>(arg: any) => GType;
/**
 * Shorthand for casting a value to it's boolean equivalent.
 *
 * @param value - the value to transform into a boolean
 *
 * @public
 */
export declare const bool: (value: unknown) => boolean;
/**
 * Check if an instance is the direct instance of the provided class.
 */
export declare const isDirectInstanceOf: <T>(instance: unknown, Constructor: AnyConstructor<T>) => instance is T;
/**
 * Predicate check that value is undefined
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isUndefined: (value: unknown) => value is undefined;
/**
 * Predicate check that value is a string
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isString: (value: unknown) => value is string;
/**
 * Predicate check that value is a number.
 *
 * Also by default doesn't include NaN as a valid number.
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isNumber: (value: unknown) => value is number;
/**
 * Predicate check that value is a function
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isFunction: (value: unknown) => value is AnyFunction<any>;
/**
 * Predicate check that value is null
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isNull: (value: unknown) => value is null;
/**
 * Predicate check that value is a class
 *
 * @deprecated Due to the current build process stripping out classes
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isClass: (value: unknown) => value is AnyConstructor<unknown>;
/**
 * Predicate check that value is boolean
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isBoolean: (value: unknown) => value is boolean;
/**
 * Predicate check that value is a symbol
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isSymbol: (value: unknown) => value is symbol;
/**
 * Helper function for Number.isInteger check allowing non numbers to be tested
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isInteger: (value: unknown) => value is number;
/**
 * Helper function for Number.isSafeInteger allowing for unknown values to be
 * tested
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isSafeInteger: (value: unknown) => value is number;
/**
 * Predicate check for whether passed in value is a plain object
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isPlainObject: (value: unknown) => value is PlainObject;
/**
 * Utility predicate check that value is either null or undefined
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isNullOrUndefined: (value: unknown) => value is null | undefined;
/**
 * Predicate check that value is an object
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isObject: (value: unknown) => value is object;
/**
 * A shorthand method for creating instance of checks.
 */
export declare const isInstanceOf: <GConstructor extends AnyConstructor<unknown>>(Constructor: GConstructor) => (value: unknown) => value is InstanceType<GConstructor>;
/**
 * Predicate check that value is a native promise
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isNativePromise: (value: unknown) => value is Promise<unknown>;
/**
 * Predicate check that value has the promise api implemented
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isPromise: (value: unknown) => value is Promise<unknown>;
/**
 * Predicate check that value is a RegExp
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isRegExp: (value: unknown) => value is RegExp;
/**
 * Predicate check that value is a date
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isDate: (value: unknown) => value is Date;
/**
 * Predicate check that value is an error
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isError: (value: unknown) => value is Error;
/**
 * Predicate check that value is a `Map`
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isMap: (value: unknown) => value is Map<unknown, unknown>;
/**
 * Predicate check that value is a `Set`
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isSet: (value: unknown) => value is Set<unknown>;
/**
 * Predicate check that value is an empty object
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isEmptyObject: (value: unknown) => value is {
    [key: string]: never;
};
/**
 * Alias the isArray method.
 */
export declare const isArray: (arg: any) => arg is any[];
/**
 * Predicate check that value is an empty array
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const isEmptyArray: (value: unknown) => value is never[];
/**
 * Capitalizes a string value.
 *
 * @param str - the string to capitalize.
 * @public
 */
export declare const capitalize: (str: string) => string;
/**
 * Removes leading and trailing whitespace from a string.
 *
 * @param str - the string to trim
 *
 * @public
 */
export declare const trim: (str: string) => string;
/**
 * Trim and conditionally capitalize string values.
 *
 * @param str - the string to format.
 *
 * @public
 */
export declare const format: (str: string) => string;
/**
 * Calls a function if defined and provides compile time type checking for the
 * passed in parameters.
 *
 * @param fn - the function to call if it exists
 * @param args - the rest of the parameters with types
 */
export declare const callIfDefined: <GFunc extends AnyFunction<any>>(fn: unknown, ...args: Parameters<GFunc>) => void;
/**
 * Finds all the regex matches for a string
 *
 * @param text - the text to check against
 * @param regexp - the regex (which should include a 'g' flag)
 *
 * @public
 */
export declare const findMatches: (text: string, regexp: RegExp) => RegExpExecArray[];
/**
 * A utility function to clean up the Operating System name.
 *
 * @param os - the OS name to clean up.
 * @param pattern - a `RegExp` pattern matching the OS name.
 * @param label - a label for the OS.
 * @returns a cleaned up Operating System name
 *
 * @public
 */
export declare const cleanupOS: (os: string, pattern?: string | undefined, label?: string | undefined) => string;
/**
 * A utility function to check whether the current browser is running on the
 * android platform.
 * @public
 */
export declare const isAndroidOS: () => boolean;
/**
 * Generate a random float between min and max. If only one parameter is
 * provided minimum is set to 0.
 *
 * @param min - the minimum value
 * @param max - the maximum value
 *
 * @public
 */
export declare const randomFloat: (min: number, max?: number | undefined) => number;
/**
 * Generate a random integer between min and max. If only one parameter is
 * provided minimum is set to 0.
 *
 * @param min - the minimum value
 * @param max - the maximum value
 *
 * @public
 */
export declare const randomInt: (min: number, max?: number | undefined) => number;
/**
 * Converts a string, including strings in camelCase or snake_case, into Start
 * Case (a variant of Title Case where all words start with a capital letter),
 * it keeps original single quote and hyphen in the word.
 *
 *   'management_companies' to 'Management Companies' 'managementCompanies' to
 *   'Management Companies' `hell's kitchen` to `Hell's Kitchen` `co-op` to
 *   `Co-op`
 *
 * @param str - the string to examine
 */
export declare const startCase: (str: string) => string;
/**
 * Returns the kebab cased form of a string.
 *
 * Taken from
 * https://github.com/angus-c/just/blob/master/packages/string-kebab-case/index.js
 * kebabCase('the quick brown fox'); // 'the-quick-brown-fox'
 * kebabCase('the-quick-brown-fox'); // 'the-quick-brown-fox'
 * kebabCase('the_quick_brown_fox'); // 'the-quick-brown-fox'
 * kebabCase('theQuickBrownFox'); // 'the-quick-brown-fox'
 * kebabCase('theQuickBrown Fox'); // 'the-quick-brown-fox'
 * kebabCase('thequickbrownfox'); // 'thequickbrownfox' kebabCase('the - quick *
 * brown# fox'); // 'the-quick-brown-fox' kebabCase('theQUICKBrownFox'); //
 * 'the-q-u-i-c-k-brown-fox'
 */
export declare const kebabCase: (str: string) => string;
/**
 * This is required since it is not exported from `fastMemomize`
 *
 * @internalRemarks
 * TODO create a pull request to fix this.
 */
export interface Memoize extends MemoizeFunc {
    strategies: {
        variadic: MemoizeFunc;
        monadic: MemoizeFunc;
    };
}
/**
 * Alias for caching function calls
 */
export declare const memoize: Memoize;
interface UniqueIdParams {
    /**
     * The prefix for the unique id
     *
     * @defaultValue ''
     */
    prefix?: string;
    /**
     * The length of the generated ID for the unique id
     *
     * @defaultValue 21
     */
    size?: number;
}
/**
 * Generate a unique id
 *
 * @param params - the destructured params
 * @returns a unique string of specified length
 *
 * @public
 */
export declare const uniqueId: ({ prefix, size }?: UniqueIdParams) => string;
/**
 * Takes a number of elements from the provided array starting from the
 * zero-index
 *
 * @param arr - the array to take from
 * @param num - the number of items to take
 *
 * @public
 */
export declare const take: <GArray extends any[]>(arr: GArray, num: number) => any[];
/**
 * Alias for picking properties from an object
 */
export declare const pick: typeof objectPick;
/**
 * Alias for excluding properties from an object
 */
export declare const omit: typeof objectOmit;
export declare const omitUndefined: (object: PlainObject) => {
    [key: string]: any;
};
/**
 * Clones a plain object using object spread notation
 *
 * @param value - the value to check
 *
 * @public
 */
export declare const clone: <GObject extends {}>(value: GObject) => GObject;
/**
 * Alias for fast deep equal
 */
export declare const isEqual: (a: any, b: any) => boolean;
/**
 * Create a unique array in a non-mutating manner
 *
 * @param array - the array which will be reduced to its unique elements
 * @param fromStart - when set to true the duplicates will be removed from the
 * beginning of the array. This defaults to false.
 * @return a new array containing only unique elements (by reference)
 *
 * @public
 */
export declare const uniqueArray: <GType>(array: GType[], fromStart?: boolean) => GType[];
/**
 * Flattens an array
 *
 * @param array
 *
 * @public
 */
export declare const flattenArray: <GType>(array: any[]) => GType[];
/**
 * Sometimes doing nothing is the best policy.
 */
export declare const noop: () => void;
/**
 * Use this to completely overwrite an object when merging.
 *
 * ```ts
 * const source = { awesome: { a: 'a' } }
 * const target = { awesome: { b: 'b' } }
 * const result = deepMerge(source, target) // => { awesome: { a: 'a', b: 'b' } }
 *
 * const overwriteTarget = { awesome: Merge.overwrite({ b: 'b' }) }
 * const overwriteResult = deepMerge(source, overwriteTarget) // => { awesome: { b: 'b' } }
 * ```
 *
 */
export declare class Merge {
    /**
     * Create an object that will completely replace the key when merging.
     *
     * @param [obj] - the object to replace the key with. When blank an empty
     * object is used.
     */
    static overwrite<GReturn = any>(obj?: PlainObject): GReturn;
    /**
     * Sets the key to undefined thus fully deleting the key.
     */
    static delete(): any;
    /**
     * This can be used to mimic any object shape.
     */
    [key: string]: any;
    private constructor();
}
/**
 * A deep merge which only merges plain objects and Arrays. It clones the object
 * before the merge so will not mutate any of the passed in values.
 *
 * To completely remove a key you can use the `Merge` helper class which
 * replaces it's key with a completely new object
 */
export declare const deepMerge: <GType = any>(...objects: (PlainObject | unknown[])[]) => GType;
interface ClampParams {
    min: number;
    max: number;
    value: number;
}
/**
 * Clamps the value to the provided range.
 */
export declare const clamp: ({ min, max, value }: ClampParams) => number;
/**
 * Get the last element of the array.
 */
export declare const last: <GType>(array: GType[]) => GType;
/**
 * Sorts an array while retaining the original order when the compare method
 * identifies the items as equal.
 *
 * `Array.prototype.sort()` is unstable and so values that are the same will
 * jump around in a non deterministic manner. Here I'm using the index as a
 * fallback. If two elements have the same priority the element with the lower
 * index is placed first hence retaining the original order.
 *
 * @param array - the array to sort
 * @param compareFn - compare the two value arguments `a` and `b` - return 0 for
 *                  equal - return number > 0 for a > b - return number < 0 for
 *                  b > a
 */
export declare const sort: <GType>(array: GType[], compareFn: (a: GType, b: GType) => number) => GType[];
/**
 * Get a property from an object or array by a string path or an array path.
 *
 * @param path - path to property
 * @param obj - object to retrieve property from
 */
export declare const get: <GReturn = any>(path: string | (string | number)[], obj: any, fallback?: any) => GReturn;
export * from 'throttle-debounce';
/**
 * Create a unique array of objects from a getter function or a property list.
 *
 * @param array - the array to extract unique values from
 * @param getValue - a getter function or a string with the path to the item
 * that is being used as a a test for uniqueness.
 * @param fromStart - when true will remove duplicates from the start rather
 * than from the end
 *
 * ```ts
 * import { uniqueBy } from '@remirror/core-helpers';
 *
 * const values = uniqueBy([{ id: 'a', value: 'Awesome' }, { id: 'a', value: 'ignored' }], item => item.id);
 * log(values) // => [{id: 'a', value: 'Awesome'}]
 *
 * const byKey = uniqueBy([{ id: 'a', value: 'Awesome' }, { id: 'a', value: 'ignored' }], 'id')
 * // Same as above
 * ```
 */
export declare const uniqueBy: <GItem = any, GKey = any>(array: GItem[], getValue: string | (string | number)[] | ((item: GItem) => GKey), fromStart?: boolean) => GItem[];
/**
 * A typesafe implementation of `Object.entries`
 *
 * Taken from
 * https://github.com/biggyspender/ts-entries/blob/master/src/ts-entries.ts
 */
export declare const entries: <GType extends object, GKey extends Extract<keyof GType, string>, GValue extends GType[GKey], GEntry extends [GKey, GValue]>(obj: GType) => GEntry[];
/**
 * A typesafe implementation of `Object.keys`
 */
export declare const keys: <GObj extends object, GKey extends Extract<keyof GObj, string>>(obj: GObj) => GKey[];
/**
 * A more lenient typed version of `Array.prototype.includes` which allow less
 * specific types to be checked.
 */
export declare const includes: <GType>(array: readonly GType[] | GType[], item: unknown, fromIndex?: number | undefined) => item is GType;
/**
 * Create a range from start to end.
 *
 * If only start is provided it creates an array of the size provided. if start
 * and end are provided it creates an array who's first position is start and
 * final position is end. i.e. `length = (end - start) + 1`
 */
export declare const range: (start: number, end?: number | undefined) => number[];
/**
 * Check that a number is within the minimum and maximum bounds of a set of
 * numbers.
 *
 * @param value - the number to test
 */
export declare const within: (value: number, ...rest: (number | null | undefined)[]) => boolean;
/**
 * Safe implementation of hasOwnProperty with typechecking.
 *
 * @remarks
 *
 * See {@link https://eslint.org/docs/rules/no-prototype-builtins}
 *
 * @param obj - the object to check
 * @param key - the property to check
 *
 * @typeParam GObj - the object type
 * @typeParam GProperty - the property which can be a string | number | symbol
 */
export declare const hasOwnProperty: <GObj extends {}, GProperty extends string | number | symbol>(obj: GObj, key: GProperty) => obj is GObj & {
    GKey: unknown;
};
//# sourceMappingURL=core-helpers.d.ts.map